// Copyright (C) 2019  Wilko Manger
// Copyright (C) 2019  Mathieu Velten
//
// This file is part of matrix-dart-sdk.
//
// matrix-dart-sdk is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// matrix-dart-sdk is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
//
// You should have received a copy of the GNU Affero General Public License
// along with matrix-dart-sdk.  If not, see <https://www.gnu.org/licenses/>.

import 'package:matrix_sdk/matrix_sdk.dart';
import 'package:matrix_sdk/src/event/room/message_event.dart';
import 'package:matrix_sdk/src/event/room/state/power_levels_change_event.dart';
import 'package:matrix_sdk/src/event/room/state/room_upgrade_event.dart';
import 'package:matrix_sdk/src/event/room/state/topic_change_event.dart';
import 'package:matrix_sdk/src/identifier.dart';
import 'package:matrix_sdk/src/util/json.dart';
import 'package:meta/meta.dart';

abstract class Event extends Identifiable<EventId> implements JsonWritable {
  String get matrixType => matrixTypeOf(this);

  @override
  final EventId id;

  final SentState sentState = SentState.sent;

  final UnsignedData unsigned;

  EventContent get content;

  Event(this.id, this.unsigned);

  factory Event.fromContent(EventContent content, RoomEventArgs args) {
    if (content is TextMessage) {
      return TextMessageEvent(content, args);
    } else if (content is AudioMessage) {
      return AudioMessageEvent(content, args);
    } else {
      return null;
    }
  }

  @override
  Map<String, dynamic> toJson() {
    return {
      'event_id': id.toJson(),
      'type': matrixType,
      'content': content?.toJson() ?? {},
      'unsigned': unsigned?.toJson(),
    };
  }

  @override
  String toString() => toJson().toString();

  // Matrix type handling
  static const Map<Type, String> matrixTypes = {
    MessageEvent: 'm.room.message',
    MemberChangeEvent: 'm.room.member',
    RedactionEvent: 'm.room.redaction',
    RoomAvatarChangeEvent: 'm.room.avatar',
    RoomNameChangeEvent: 'm.room.name',
    RoomCreationEvent: 'm.room.create',
    RoomUpgradeEvent: 'm.room.tombstone',
    TopicChangeEvent: 'm.room.topic',
    PowerLevelsChangeEvent: 'm.room.power_levels'
  };

  /// Get the type name as used by Matrix for an [Event]'s [Type], or of
  /// an [Event] instance, for example: `m.room.message`.
  ///
  /// [input] can either be an [Event] instance or a [Type] of a certain
  /// [Event].
  ///
  /// If the positional [input] argument is given instead of the generic one,
  /// it prefers that. The [input] can either be an [Event] instance or
  /// an [Event] type.
  ///
  /// **WARNING:** If only [T] is given, it will only work correctly on
  /// base types of events (e.g. [MessageEvent] but not [TextMessageEvent].
  static String matrixTypeOf<T extends Event>([dynamic input]) {
    var event;
    var type = T;
    if (input is Event) {
      event = input;
    } else if (input is Type) {
      type = input;
    }

    if (event is MessageEvent || type == MessageEvent) {
      return matrixTypes[MessageEvent];
    } else if (event is MemberChangeEvent || type == MemberChangeEvent) {
      return matrixTypes[MemberChangeEvent];
    } else if (event is RedactionEvent || type == RedactionEvent) {
      return matrixTypes[RedactionEvent];
    } else if (event is RoomAvatarChangeEvent ||
        type == RoomAvatarChangeEvent) {
      return matrixTypes[RoomAvatarChangeEvent];
    } else if (event is RoomNameChangeEvent || type == RoomNameChangeEvent) {
      return matrixTypes[RoomNameChangeEvent];
    } else if (event is RoomCreationEvent || type == RoomCreationEvent) {
      return matrixTypes[RoomCreationEvent];
    } else if (event is RoomUpgradeEvent || type == RoomUpgradeEvent) {
      return matrixTypes[RoomUpgradeEvent];
    } else if (event is TopicChangeEvent || type == TopicChangeEvent) {
      return matrixTypes[TopicChangeEvent];
    } else if (event is PowerLevelsChangeEvent ||
        type == PowerLevelsChangeEvent) {
      return matrixTypes[PowerLevelsChangeEvent];
    } else {
      return null;
    }
  }

  /// Get the base [Type] of an [Event] associated by the given [matrixType].
  static Type typeOf(String matrixType) => matrixTypes.entries
      .firstWhere((entry) => entry.value == matrixType, orElse: () => null)
      ?.key;
}

enum SentState { unsent, failed, sent }

abstract class EventContent {
  Map<String, dynamic> toJson() => {};
}

class UnsignedData {
  final String transactionId;

  UnsignedData({@required this.transactionId});

  dynamic toJson() => {'transaction_id': transactionId};
}
