// Copyright (C) 2019  Wilko Manger
//
// This file is part of matrix-dart-sdk.
//
// matrix-dart-sdk is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// matrix-dart-sdk is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
//
// You should have received a copy of the GNU Affero General Public License
// along with matrix-dart-sdk.  If not, see <https://www.gnu.org/licenses/>.

import 'package:matrix_sdk/src/event/room/room_event.dart';
import 'package:matrix_sdk/src/event/room/state/state_event.dart';
import 'package:matrix_sdk/src/identifier.dart';

import '../../event.dart';

class PowerLevelsChangeEvent extends StateEvent {
  @override
  final PowerLevelsChange content;

  PowerLevelsChangeEvent(this.content, RoomEventArgs args)
      : super(args, stateKey: '');

  factory PowerLevelsChangeEvent.fromJson(RoomEventArgs args, dynamic json) {
    if (!hasStateKey(json)) {
      return null;
    }

    final content = json['content'];

    final int ban = content['ban'] ?? 50;
    final int invite = content['invite'] ?? 50;
    final int kick = content['kick'] ?? 50;
    final int redact = content['redact'] ?? 50;

    final int eventsDefault = content['events_default'] ?? 0;
    final filteredEvents = content['events']
      ..removeWhere((type, powerLevel) => Event.typeOf(type) == null);

    final Map<Type, int> events = filteredEvents.map((eventType, powerLevel) {
      return MapEntry<Type, int>(Event.typeOf(eventType), powerLevel);
    }).cast<Type, int>();

    final int stateDefault = content['state_default'] ?? 50;

    final int userDefault = content['users_default'] ?? 0;
    final Map<UserId, int> users = content['users']
        .map((userId, powerLevel) => MapEntry(UserId(userId), powerLevel))
        .cast<UserId, int>();

    final roomNotifications = json.containsKey('notifications')
        ? json['notifications']['room'] ?? 50
        : 50;

    return PowerLevelsChangeEvent(
      PowerLevelsChange(
        banLevel: ban,
        inviteLevel: invite,
        kickLevel: kick,
        redactLevel: redact,
        stateEventsDefaultLevel: stateDefault,
        eventsDefaultLevel: eventsDefault,
        eventLevels: events,
        userDefaultLevel: userDefault,
        userLevels: users,
        roomNotificationLevel: roomNotifications,
      ),
      args,
    );
  }
}

class PowerLevelsChange extends StateEventContent {
  final int banLevel;
  final int inviteLevel;
  final int kickLevel;
  final int redactLevel;

  final int stateEventsDefaultLevel;
  final int eventsDefaultLevel;
  final Map<Type, int> eventLevels;

  final int userDefaultLevel;
  final Map<UserId, int> userLevels;

  final int roomNotificationLevel;

  PowerLevelsChange({
    this.banLevel = 50,
    this.inviteLevel = 50,
    this.kickLevel = 50,
    this.redactLevel = 50,
    this.stateEventsDefaultLevel = 50,
    this.eventsDefaultLevel = 0,
    this.eventLevels,
    this.userDefaultLevel = 0,
    this.userLevels,
    this.roomNotificationLevel = 50,
  });

  @override
  Map<String, dynamic> toJson() => super.toJson()
    ..addAll({
      'ban': banLevel,
      'invite': inviteLevel,
      'kick': kickLevel,
      'redact': redactLevel,
      'state_default': stateEventsDefaultLevel,
      'events_default': eventsDefaultLevel,
      'events': eventLevels
          .map((type, power) => MapEntry(Event.matrixTypeOf(type), power)),
      'users_default': userDefaultLevel,
      'users':
          userLevels.map((userId, power) => MapEntry(userId.toString(), power)),
      'notifications': {
        'room': roomNotificationLevel,
      },
    });
}
