// Copyright (C) 2019  Wilko Manger
// Copyright (C) 2019  Mathieu Velten
//
// This file is part of matrix-dart-sdk.
//
// matrix-dart-sdk is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// matrix-dart-sdk is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
//
// You should have received a copy of the GNU Affero General Public License
// along with matrix-dart-sdk.  If not, see <https://www.gnu.org/licenses/>.

import 'dart:convert';

import 'package:chopper/chopper.dart';
import 'package:http/http.dart' as http;
import 'package:matrix_sdk/src/api/client.dart';
import 'package:matrix_sdk/src/api/media.dart';
import 'package:matrix_sdk/src/exception.dart';

import 'package:matrix_sdk/src/identifier.dart';
import 'package:matrix_sdk/src/store.dart';
import 'package:matrix_sdk/src/user.dart';
import 'package:matrix_sdk/src/util/chopper.dart';

import 'package:kiwi/kiwi.dart';
import 'package:url/url.dart';
import 'package:meta/meta.dart';

/// Represents a Matrix homeserver. Also used as the main entry point
/// of the SDK.
class Homeserver {
  final ChopperClient _chopper;
  // TODO: Make lazy
  ClientService get _client => _chopper.getService();
  MediaService get _media => _chopper.getService();

  final Url url;

  /// Dependency injection scope.
  /// @internal
  String scope;

  /// Returns a homeserver based on `.well-known` info.
  /// If there is no well-known info (404), just use [url].
  ///
  /// Queries `$url/.well-known/matrix/client` for information.
  ///
  /// Throws [WellKnownException] on failure.
  static Future<Homeserver> fromWellKnown(Url url) async {
    final wellKnownUrl = url.resolve('/.well-known/matrix/client');

    http.Client client = http.Client();
    final response = await client.get(wellKnownUrl);

    if (response.statusCode == 404) {
      return Homeserver(url);
    }

    if (response.statusCode != 200) {
      throw WellKnownFailPromptException('Response status code is not 200');
    }

    if (response.body == null || response.body.isEmpty) {
      throw WellKnownFailPromptException('Response body is null or empty');
    }

    var body;
    try {
      body = json.decode(response.body);
    } on FormatException {
      throw WellKnownFailPromptException('Response body is not valid JSON');
    }

    final homeserverInfo = body['m.homeserver'];
    if (homeserverInfo == null) {
      throw WellKnownFailPromptException('m.homeserver key is missing');
    }

    final baseUrl = homeserverInfo['base_url'];
    if (baseUrl == null) {
      throw WellKnownFailPromptException('base_url key is missing');
    }

    try {
      final url = Url.parse(baseUrl);
      return Homeserver(url);
    } on FormatException {
      throw WellKnownFailErrorException('base_url is not valid');
    }

    // TODO: Validation
  }

  Homeserver(this.url) : _chopper = createChopper(url: url) {
    scope = url.toString() + hashCode.toString();

    Container().registerInstance(this, name: scope);
    Container().registerInstance(_client, name: scope);
    Container().registerInstance(_media, name: scope);
  }

  /// Logs in to this homeserver and returns a [LocalUser], which can be
  /// used for further operations.
  ///
  /// [user]: Can be either a [UserId] or just a [Username].
  /// [password]: Password of the given [user].
  /// [store]: Store where all user data (rooms, events, etc) will be stored.
  Future<LocalUser> login(
    UserIdentifier user,
    String password, {
    Store store,
  }) async {
    final response = await _client.login(json.encode({
      'type': 'm.login.password',
      'identifier': user.toIdentifierJson(),
      'password': password,
    }));

    final body = json.decode(response.body);

    String accessToken = body['access_token'];
    final userId = UserId(body['user_id']);

    final localUser = LocalUser(
      id: userId,
      store: store,
      homeserver: this,
      accessToken: accessToken,
    );

    await store?.open(user: localUser);

    await store?.setUserId(userId);
    await store?.setHomeserver(this);
    await store?.setAccessToken(accessToken);

    return localUser;
  }

  /// Checks whether the given [username] is available for
  /// registration on this [Homeserver].
  Future<bool> isUsernameAvailable(Username username) async {
    var response;

    try {
      response = await _client.registerAvailable(username.toString());
    } on UsernameInUseException {
      return false;
    }

    switch (response.statusCode) {
      case 200:
        return true;
      default:
        return false;
    }
  }

  /// Download content via this [Homeserver].
  Future<Stream<List<int>>> download(Url url) async {
    if (url == null) {
      throw ArgumentError.value(url, 'url', 'Must not be null');
    }

    if (url.scheme != 'mxc') {
      throw ArgumentError.value(url, 'url', 'Must be an mxc URL');
    }

    final server = url.authority;
    final mediaId = url.pathSegments[0];

    final response = await _media.download(server, mediaId);
    return response.body;
  }

  /// Download a thumbnail via this [Homeserver].
  Future<Stream<List<int>>> downloadThumbnail(
    Url url, {
    @required int width,
    @required int height,
    ResizeMethod resizeMethod = ResizeMethod.scale,
  }) async {
    if (url == null) {
      throw ArgumentError.value(url, 'url', 'Must not be null');
    }

    if (url.scheme != 'mxc') {
      throw ArgumentError.value(url, 'url', 'Must be an mxc URL');
    }

    final method = resizeMethod == ResizeMethod.scale ? 'scale' : 'crop';

    final server = url.authority;
    final mediaId = url.pathSegments[0];

    final response = await _media.thumbnail(
      server,
      mediaId,
      width,
      height,
      method,
    );
    return response.body;
  }
}

enum ResizeMethod {
  crop,
  scale,
}

/// Extracting information from `.well-known` failed. See
/// [message] for details.
abstract class WellKnownException implements Exception {
  final String message;

  WellKnownException(this.message);

  @override
  String toString() => '$runtimeType: $message';
}

/// Extracting information from `.well-known` failed. See
/// [message] for details.
///
/// Clients should `FAIL_PROMPT`, as per the spec.
class WellKnownFailPromptException extends WellKnownException {
  WellKnownFailPromptException(String message) : super(message);
}

/// Extracting information from `.well-known` failed. See
/// [message] for details.
///
/// Clients should `FAIL_ERROR`, as per the spec.
class WellKnownFailErrorException extends WellKnownException {
  WellKnownFailErrorException(String message) : super(message);
}
