// Copyright (C) 2019  Wilko Manger
// Copyright (C) 2019  Mathieu Velten
//
// This file is part of matrix-dart-sdk.
//
// matrix-dart-sdk is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// matrix-dart-sdk is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
//
// You should have received a copy of the GNU Affero General Public License
// along with matrix-dart-sdk.  If not, see <https://www.gnu.org/licenses/>.

import 'dart:convert';

import 'package:kiwi/kiwi.dart';
import 'package:matrix_sdk/src/api/client.dart';
import 'package:matrix_sdk/src/event/event.dart';
import 'package:matrix_sdk/src/event/room/state/member_change_event.dart';
import 'package:matrix_sdk/src/event/room/state/room_creation_event.dart';
import 'package:matrix_sdk/src/event/room/state/room_upgrade_event.dart';
import 'package:matrix_sdk/src/event/room/state/topic_change_event.dart';
import 'package:matrix_sdk/src/event/timeline.dart';
import 'package:matrix_sdk/src/event/room/redaction_event.dart';
import 'package:matrix_sdk/src/event/room/room_event.dart';
import 'package:matrix_sdk/src/event/room/state/room_avatar_change_event.dart';
import 'package:matrix_sdk/src/event/room/state/room_name_change_event.dart';
import 'package:matrix_sdk/src/identifier.dart';
import 'package:matrix_sdk/src/members.dart';
import 'package:matrix_sdk/src/user.dart';
import 'package:chopper/chopper.dart';
import 'package:matrix_sdk/src/util/random.dart';
import 'package:meta/meta.dart';
import 'package:url/url.dart';

import '../store.dart';

class Room extends Identifiable<RoomId> {
  @override
  final RoomId id;

  Timeline _timeline;
  Timeline get timeline => _timeline;
  final Members members;

  //// @internal
  String previousBatch;

  final String _scope;
  ClientService _client;

  RoomStateEvents _stateEvents = RoomStateEvents();

  /// The latest [StateEvent]s of this room.
  RoomStateEvents get stateEvents => _stateEvents;

  User _creator;

  /// [UserId] of the user that created this room.
  User get creator => _creator;

  RoomId _replacementId;

  /// The [RoomId] of the room that replaced this room.
  RoomId get replacementId => _replacementId;

  /// Whether this room is upgraded to another room.
  ///
  /// See [replacementId] to see which room this was upgraded.
  bool get isUpgraded => replacementId != null;

  RoomId _replacesId;

  /// The [RoomId] of the room that was replaced by this room.
  RoomId get replacesId => _replacesId;

  /// Whether this room is a replacement for another room.
  ///
  /// See [replacesId] to see which room was replaced by this room.
  bool get isReplacement => replacesId != null;

  String _topic;

  String get topic => _topic;

  List<User> _typingUsers = List();

  List<User> get typingUsers => _typingUsers;

  final LocalUser _me;

  final Store _store;

  Room({
    @required this.id,
    @required String scope,
    int memberCount,
    String name,
    Url avatarUrl,
    String topic,
    User directUser,
    String previousBatch,
    User creator,
    RoomId replacementId,
    RoomId replacesId,
  })  : _scope = scope,
        _name = name,
        _avatarUrl = avatarUrl,
        _topic = topic,
        _directUser = directUser,
        previousBatch = previousBatch,
        _creator = creator,
        _replacementId = replacementId,
        _replacesId = replacesId,
        _me = Container().resolve(scope),
        _store = Container().resolve(scope),
        members = Members(id, memberCount, scope) {
    _timeline = Timeline(this, scope);
    _client = Container().resolve(_scope);
  }

  String _name;
  String get name => _name;

  Url _avatarUrl;
  Url get avatarUrl => _avatarUrl;

  bool get isDirect => directUser != null;

  User _directUser;

  /// The other user of the direct chat. Is null if [isDirect] is false.
  User get directUser => _directUser;

  /// @internal
  Future<Iterable<RoomEvent>> retrieveRemoteEvents(int limit) async {
    final response = await _client.roomMessages(
      roomId: id.toString(),
      limit: limit,
      from: previousBatch,
      filter: json.encode({
        'lazy_load_members': true,
      }),
    );

    final body = json.decode(response.body);

    previousBatch = body['end'];
    await _store.setRooms([this]);

    final context = ProcessContext();

    if (body.containsKey('state')) {
      await _processEvents(body['state'], context);
    }

    // TODO: When most event types are supported, don't check if they're
    // null anymore (null is now returned if the event type is unsupported)
    final processed = (await _processEvents(body['chunk'].reversed, context))
        .where((event) => event != null);

    return processed;
  }

  /// @internal
  Future<void> processSync(Map<String, dynamic> response) async {
    // Only set previous batch from sync when none is set
    if (previousBatch == null) {
      previousBatch = response['timeline']['prev_batch'];
    }

    final summary = response['summary'];
    if (summary != null) {
      members.count = summary['m.joined_member_count'] ?? members.count;
    }

    final context = ProcessContext();

    await _processEvents(response['state']['events'], context);

    final processed = await _processEvents(
      response['timeline']['events'],
      context,
    );

    await timeline.addAll(processed);
  }

  Future<List<RoomEvent>> _processEvents(
      Iterable<dynamic> rawEvents, ProcessContext context) async {
    RoomCreationEvent creationEvent;
    final processedEvents = List<RoomEvent>();
    for (final rawEvent in rawEvents) {
      final event = await RoomEvent.fromJson(
        RoomEventArgs(room: this),
        rawEvent,
        context: context,
      );

      processedEvents.add(event);

      if (event is RedactionEvent) {
        final redactedId = event.redacts;

        final original = await timeline[redactedId];
        if (original != null && original is! RedactedEvent) {
          await timeline.remove(redactedId);

          await timeline.addAll([
            RedactedEvent.from(
              redaction: event,
              original: original,
            )
          ]);
        }
      } else if (event is RoomNameChangeEvent) {
        var name = event.content.name;
        // Set to null if name is empty
        if (name != null && name.isEmpty) {
          name = null;
        }

        _name = name;

        _stateEvents = _stateEvents.copyWith(nameChangeEvent: event);
      } else if (event is RoomAvatarChangeEvent) {
        _avatarUrl = event.content.url;
      } else if (event is RoomCreationEvent) {
        creationEvent = event;
      } else if (event is RoomUpgradeEvent) {
        _replacementId = event.content.replacementRoomId;
      } else if (event is TopicChangeEvent) {
        _topic = event.content.topic;
      } else if (event is InviteEvent && event.content.isDirect) {
        // Check whether there are only two people in the
        // room
        if (members.count <= 2) {
          if (!event.content.subject.isIdenticalTo(_me)) {
            _directUser = event.content.subject;
          } else if (!event.sender.isIdenticalTo(_me)) {
            _directUser = event.sender;
          }
        }
      } else if (event is JoinEvent && this.isDirect) {
        // If someone joins who is not me or the direct user,
        // this room is no longer direct
        final subject = event.content.subject;
        if (!subject.isIdenticalTo(_me) && !subject.isIdenticalTo(directUser)) {
          _directUser = null;
        }
      }
    }

    await members.addAll(context.users);

    // Do this after the members have been added
    // so the correct user state can be set
    if (creationEvent != null) {
      _creator = await members.getStateOf(
        creationEvent.content.creatorId,
        at: creationEvent.time,
      );

      if (_creator == null) {
        _creator = User(
          id: creationEvent.content.creatorId,
          state: UserState(
            roomId: id,
            since: creationEvent.time,
          ),
        );

        await members.addAll([_creator]);
      }

      _replacesId = creationEvent.content.previousRoomId;
    }

    return processedEvents;
  }

  Future<void> processUserAccountData(Map<String, dynamic> raw) async {
    final events = raw['events'] as List<dynamic>;

    if (_directUser == null) {
      final event = events.firstWhere(
        (event) => event['type'] == 'm.direct',
        orElse: () => null,
      );

      if (event != null) {
        final content = event['content'] as Map<String, dynamic>;

        await content.forEach((userId, roomIds) async {
          if (UserId.isValidFullyQualified(userId) &&
              roomIds.contains(id.toString())) {
            // Check whether there are only two people in the
            // room
            if (members.count <= 2) {
              _directUser = await members[UserId(userId)];
            }
          }
        });
      }
    }
  }

  Future<void> processEphemeral(Map<String, dynamic> raw) async {
    final events = raw['ephemeral']['events'] as List<dynamic>;

    final rawTypingEvent = events.firstWhere(
      (rawEvent) => rawEvent['type'] == 'm.typing',
      orElse: () => null,
    );

    if (rawTypingEvent == null) return;

    final typingUsers = (await Future.wait(rawTypingEvent['content']['user_ids']
            .map((id) => Future.sync(() => members[UserId(id)]))
            .cast<Future<User>>()))
        .cast<User>();

    _typingUsers = typingUsers;
  }
}

class ProcessContext {
  final Map<EventId, RoomEvent> timeline = Map();

  /// User which is currently the correct one.
  final Map<UserId, User> currentUsers = Map();

  /// All user states.
  final List<User> users = List();

  void setUser(User user) {
    users.add(user);
    currentUsers[user.id] = user;
  }
}

class JoinedRoom extends Room {
  JoinedRoom({
    @required RoomId id,
    @required String scope,
    int memberCount,
    String name,
    Url avatarUrl,
    String topic,
    User directUser,
    String previousBatch,
    User creator,
    RoomId replacementId,
    RoomId replacesId,
  }) : super(
          id: id,
          scope: scope,
          memberCount: memberCount,
          name: name,
          avatarUrl: avatarUrl,
          topic: topic,
          directUser: directUser,
          previousBatch: previousBatch,
          replacementId: replacementId,
          replacesId: replacesId,
          creator: creator,
        );

  /// Send an [Event] to the [Room].
  /// Returns the up to date [SentState] of the sent event. (Not
  /// including [SentState.received].
  Stream<SentState> send(EventContent content, {String transactionId}) async* {
    if (transactionId == null) {
      transactionId = randomString();
    }

    final eventArgs = RoomEventArgs(
      id: EventId(transactionId),
      room: this,
      time: DateTime.now(),
      sender: User(
        id: Container().resolve<LocalUser>(_scope).id,
      ),
      unsigned: UnsignedData(transactionId: transactionId),
      sentState: SentState.unsent,
    );
    final event = Event.fromContent(content, eventArgs);

    await timeline.addAll([event]);
    yield SentState.unsent;

    var sent = false;
    while (!sent) {
      Response response;

      try {
        response = await _client.send(
          roomId: id.toString(),
          eventType: event.matrixType,
          txnId: transactionId,
          content: json.encode(content.toJson()),
        );

        final body = json.decode(response.body);
        final eventId = EventId(body['event_id']);

        final sentEvent = Event.fromContent(
          content,
          eventArgs.copyWith(
            id: eventId,
            sentState: SentState.sent,
          ),
        );

        sent = true;
        await timeline.addAll([sentEvent]);
        yield SentState.sent;
      } on Exception {
        yield SentState.failed;
        continue;
      }
    }
  }

  /// Notify whether the user is typing.
  ///
  /// If [typing] is true, a [timeout] can be specified
  /// to notify how long the server should assume that the user is
  /// typing.
  Future<void> setIsTyping(
    bool typing, {
    Duration timeout = const Duration(seconds: 30),
  }) async {
    final Map<String, dynamic> body = {
      'typing': typing,
    };

    if (typing) {
      body['timeout'] = timeout.inMilliseconds;
    }

    await _client.typing(
      roomId: id.toString(),
      userId: _me.id.toString(),
      body: json.encode(body),
    );
  }
}

class LeftRoom extends Room {
  LeftRoom({
    @required RoomId id,
    @required String scope,
    int memberCount,
    String name,
    Url avatarUrl,
    String topic,
    User directUser,
    String previousBatch,
    User creator,
    RoomId replacementId,
    RoomId replacesId,
  }) : super(
          id: id,
          scope: scope,
          memberCount: memberCount,
          name: name,
          avatarUrl: avatarUrl,
          topic: topic,
          directUser: directUser,
          previousBatch: previousBatch,
          creator: creator,
          replacementId: replacementId,
          replacesId: replacesId,
        );
}

/// Class that contains the most recent state events of each type.
class RoomStateEvents {
  final RoomNameChangeEvent nameChange;
  final RoomAvatarChangeEvent avatarChange;
  final RoomCreationEvent creation;
  final RoomUpgradeEvent upgrade;

  RoomStateEvents({
    this.nameChange,
    this.avatarChange,
    this.creation,
    this.upgrade,
  });

  RoomStateEvents copyWith({
    RoomNameChangeEvent nameChangeEvent,
    RoomAvatarChangeEvent avatarChangeEvent,
    RoomCreationEvent creationEvent,
    RoomUpgradeEvent upgradeEvent,
  }) {
    return RoomStateEvents(
      nameChange: nameChangeEvent ?? this.nameChange,
      avatarChange: avatarChangeEvent ?? this.avatarChange,
      creation: creationEvent ?? this.creation,
      upgrade: upgradeEvent ?? this.upgrade,
    );
  }
}
