// Copyright (C) 2019  Wilko Manger
// Copyright (C) 2019  Mathieu Velten
//
// This file is part of matrix-dart-sdk.
//
// matrix-dart-sdk is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// matrix-dart-sdk is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
//
// You should have received a copy of the GNU Affero General Public License
// along with matrix-dart-sdk.  If not, see <https://www.gnu.org/licenses/>.

import 'dart:async';

import 'package:kiwi/kiwi.dart';
import 'package:matrix_sdk/src/identifier.dart';
import 'package:matrix_sdk/src/store.dart';
import 'package:meta/meta.dart';

/// [Identifiable] objects that can possibly be stored.
///
/// All methods that return a [FutureOr], will return a
/// [Future] only if items need to be retrieved from the [store] or
/// remotely. Subsequent calls (where the same items are retrieved)
/// will be from memory and therefore not a [Future].
abstract class Storables<I extends Id, T extends Identifiable<I>> {
  final Map<I, T> _memory = Map();

  final Store store;

  Storables(String scope) : store = Container().resolve(scope);

  @protected
  Iterable<T> getFromMemory() => _memory.values;

  @protected
  void setInMemory(Iterable<T> items) {
    _memory.addEntries((items.map((item) => MapEntry(item.id, item))));
  }

  @protected
  void setInMemoryIfAbsent(T item, T ifAbsent()) {
    _memory.putIfAbsent(item.id, ifAbsent);
  }

  @protected
  void removeInMemory(I id) {
    _memory.remove(id);
  }

  @protected
  Future<T> getSingleFromStore(I id);

  @protected
  Future<Iterable<T>> getFromStore({T from, int count});

  @protected
  Future<int> getSavedInStoreCount();

  @protected
  Future<void> setInStore(Iterable<T> value);

  @protected
  Future<void> removeInStore(I id) => null;

  @protected
  Future<Iterable<T>> getFromRemote({int count});

  @protected
  Iterable<T> sort(Iterable<T> items) => items;

  FutureOr<T> operator [](I id) {
    // If the room is in memory, return that one
    final fromMemory = _memory[id];
    if (fromMemory != null) {
      return fromMemory;
    }

    // Look in the store
    return getSingleFromStore(id);
  }

  FutureOr<Iterable<T>> get({int upTo, bool allowRemote = true}) {
    if (upTo != null) {
      return paginate(
        page: 0,
        pageSize: upTo,
        allowRemote: allowRemote,
      );
    } else {
      return getSavedInStoreCount().then((savedInStoreCount) {
        return paginate(
          page: 0,
          pageSize: savedInStoreCount,
          allowRemote: allowRemote,
        );
      });
    }
  }

  // TODO: Only sort when saving instead of getting
  FutureOr<Iterable<T>> paginate({
    @required int page,
    int pageSize = 30,
    bool allowRemote = true,
  }) {
    int start = page * pageSize;
    int end = (page + 1) * pageSize;

    final fullMemory = sort(getFromMemory());

    final memory = fullMemory.skip(start).take(pageSize);

    // TODO: Check for RoomCreationEvent (cleanly)
    if (memory.length >= pageSize) {
      return memory;
    }

    T last;
    try {
      last = fullMemory.last;
    } on StateError {
      last = null;
    }

    // Not enough items have been retrieved for the page,
    // get from store
    start += memory.length;
    return getFromStore(
      from: last,
      count: end - start,
    ).then((store) async {
      final combined = sort(memory.followedBy(store));
      setInMemory(store);

      if (combined.length >= pageSize || !allowRemote) {
        return combined;
      }

      // Not enough items have been retrieved for the page,
      // get from remote
      start += store.length;
      var remote = sort(await getFromRemote(count: end - start));

      setInMemory(remote);
      await setInStore(remote);

      T from;
      try {
        from = combined.last;
      } on StateError {
        from = null;
      }

      if (from != null) {
        var passedFrom = false;

        final filteredRemote = List<T>();
        for (final item in remote) {
          if (passedFrom) {
            filteredRemote.add(item);
          }

          if (item == from) {
            passedFrom = true;
          }
        }

        // If the filtered remote is empty, assume that the remote list
        // seamlessly joins with `combined`, otherwise
        // let remote be filtered
        if (filteredRemote.isNotEmpty) {
          remote = filteredRemote;
        }
      }

      return sort(combined.followedBy(remote));
    });
  }

  Future<void> addAll(Iterable<T> values) async {
    final nonNullValues = values.where((value) => value != null);

    setInMemory(nonNullValues);

    await setInStore(nonNullValues);
  }

  Future<void> remove(I id) async {
    removeInMemory(id);
    await removeInStore(id);
  }
}
